/* Copyright 2015 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define MAX(x,y) ((x) > (y) ? (x) : (y))

#define MAX_NEIGHBOURS 6
struct vertex {
  int x;
  int y;
  int num_neighbours;
  int neighbours[MAX_NEIGHBOURS];
  int distance;
};

/*
   (x+y*sqrt3i)*b == (3*x-3*y)/2 + (x+3*y)/2*sqrt3i
 */
#define MUL(x,y)                        \
  do {                                  \
    int MUL__x = (x);                   \
    (x) = (3*((x)-(y))) >> 1;           \
    (y) = (MUL__x + 3*(y)) >> 1;        \
  } while (0)

#define XY_ROTATE_PLUS60(x,y)                   \
  do {                                          \
    int XY_ROTATE_PLUS60__x = (x);              \
    (x) = ((x) - 3*(y)) >> 1;                   \
    (y) = ((XY_ROTATE_PLUS60__x) + (y)) >> 1;   \
  } while (0)

/*
  (x+y*sqrt3i)*w3 == (-x - 3*y)/2 + (x-y)/2 * sqrt3i
 */
#define XY_ROTATE_PLUS120(x,y)                  \
  do {                                          \
    int XY_ROTATE_PLUS120__x = (x);             \
    (x) = (- ((x) + 3*(y))) >> 1;               \
    (y) = ((XY_ROTATE_PLUS120__x)-(y)) >> 1;    \
  } while (0)

void *
malloc_or_die (size_t len)
{
  void *p = malloc(len);
  if (! p) {
    fprintf(stderr, "Cannot malloc %u\n", len);
    abort();
  }
  return p;
}

struct vertex *
find_xy (struct vertex *vp, int vplen, int x, int y)
{
  struct vertex *endp = vp + vplen;
  struct vertex *p;
  for (p = vp; p < endp; p++) {
    if (p->x == x && p->y == y) {
      return p;
    }
  }
  return NULL;
}

void
show (struct vertex *vp, int vplen)
{
  int i,j;
  for (i = 0; i < vplen; i++) {
    printf("%d at %d,%d neighbours[%d] ",
           i, vp[i].x, vp[i].y, vp[i].num_neighbours);
    for (j = 0; j < vp[i].num_neighbours; j++) {
      printf(" %d", vp[i].neighbours[j]);
    }
    printf("\n");
  }
}

/*
      7---6
       \
    4---5
     \ / \
      2---3
       \
    0---1
        14  15

      12  13

        10  11
         
       8   9   6---7
                \
        11   4---5
              \ / \
           .   2---3
                \
             0---1


 */
void
unfold (struct vertex **vpp, int *vplenp)
{
  /* printf("unfold\n"); */

  struct vertex *vp = *vpp;
  int vplen = *vplenp;
  struct vertex *endp = vp + vplen;

  struct vertex *new_vp = malloc_or_die(sizeof(*vp) * vplen * 3);
  int new_vplen = vplen;
  memcpy (new_vp, vp, sizeof(*vp) * vplen);
  struct vertex *new_endp = new_vp + new_vplen;
  /* printf("vp %p  new_vp %p\n", vp, new_vp); */

  struct vertex *hip = endp-1;
  int x_hi = hip->x;
  int y_hi = hip->y;
  struct vertex *p;
  int from;

  /* digit-1 */
  for (p = vp; p < endp; p++) {
    int x = p->x;
    int y = p->y;
    XY_ROTATE_PLUS120(x,y);
    x += x_hi;
    y += y_hi;
    /* printf(" digit-1  %d,%d -> %d,%d\n", p->x,p->y, x,y); */

    struct vertex *unfoldp = find_xy(new_vp,new_vplen, x,y);
    if (unfoldp == NULL) {
      unfoldp = new_endp++;
      new_vplen++;
      unfoldp->x = x;
      unfoldp->y = y;
      unfoldp->num_neighbours = 0;
      /* printf(" digit-1 new vertex %d at %d,%d\n", unfoldp-new_vp, x,y); */
    }
    p->distance = unfoldp - new_vp;
    /* printf(" unfold of %d is %d\n", p-vp, p->distance); */
  }

  for (from = 0; from < vplen; from++) {
    /* printf(" from %d has %d neighbours\n", from, vp[from].num_neighbours); */
    int unfold_from = vp[from].distance;
    int i;
    for (i = 0; i < vp[from].num_neighbours; i++) {
      int to = vp[from].neighbours[i];
      int unfold_to = vp[to].distance;
      /* printf("  edge %d to %d becomes %d to %d\n", */
      /*        from, to, unfold_from, unfold_to); */

      struct vertex *fromp = new_vp + unfold_from;
      if (fromp->num_neighbours >= 6) {
        fprintf(stderr, "too many neighbours\n");
        abort();
      }
      fromp->neighbours[fromp->num_neighbours++] = unfold_to;
    }
  }

  /* digit-2 */
  int mx = x_hi, my = y_hi;
  XY_ROTATE_PLUS60(mx,my);
  for (p = vp; p < endp; p++) {
    int x = p->x;
    int y = p->y;
    x += mx;
    y += my;
    /* printf(" digit-2  %d,%d -> %d,%d\n", p->x,p->y, x,y); */

    struct vertex *unfoldp = find_xy(new_vp,new_vplen, x,y);
    if (unfoldp == NULL) {
      unfoldp = new_endp++;
      new_vplen++;
      unfoldp->x = x;
      unfoldp->y = y;
      unfoldp->num_neighbours = 0;
      /* printf(" digit-2 new vertex %d at %d,%d\n", unfoldp-new_vp, x,y); */
    }
    p->distance = unfoldp - new_vp;
    /* printf(" unfold of %d is %d\n", p-vp, p->distance); */
  }

  for (from = 0; from < vplen; from++) {
    /* printf(" from %d has %d neighbours\n", from, vp[from].num_neighbours); */
    int unfold_from = vp[from].distance;
    int i;
    for (i = 0; i < vp[from].num_neighbours; i++) {
      int to = vp[from].neighbours[i];
      int unfold_to = vp[to].distance;
      /* printf("  edge %d to %d becomes %d to %d\n", */
      /*        from, to, unfold_from, unfold_to); */

      struct vertex *fromp = new_vp + unfold_from;
      if (fromp->num_neighbours >= MAX_NEIGHBOURS) {
        fprintf(stderr, "too many neighbours digit-2, edge %d--%d  unfolded %d--%d\n",
                from,to,
                unfold_from,unfold_to);
        abort();
      }
      fromp->neighbours[fromp->num_neighbours++] = unfold_to;
    }
  }


  *vpp = new_vp;
  *vplenp = new_vplen;
  free(vp);
}

/* Dijkstra's algorithm updating minimum across each edge. */
int
distance (struct vertex *vp, int vplen, int v1, int v2)
{
  /* printf("distance %d to %d\n", v1, v2); */

  struct vertex *endp = vp + vplen;
  {
    struct vertex *p;
    for (p = vp; p < endp; p++) {
      p->distance = INT_MAX;
    }
  }
  vp[v1].distance = 0;

  int updated;
  do {
    updated = 0;

    struct vertex *fromp;
    for (fromp = vp; fromp < endp; fromp++) {
      int new_distance = fromp->distance;
      if (new_distance == INT_MAX) { continue; }
      new_distance++;
      /* printf(" at %d new_distance %d\n", fromp-vp, new_distance); */

      int i;
      int num_neighbours = fromp->num_neighbours;
      for (i = 0; i < num_neighbours; i++) {
        int to = fromp->neighbours[i];
        /* printf(" to %d\n", to); */
        struct vertex *top = vp + to;
        int to_distance = top->distance;
        /* printf(" to_distance %d\n", to_distance); */
        int bigger = to_distance > new_distance;
        updated += bigger;
        top->distance = (bigger ? new_distance : to_distance);
      }
    }
    /* printf("  updated %d\n", updated); */
  } while (updated);
  /* printf("  is %d\n", vp[v2].distance); */

  return(vp[v2].distance);
}

int
main (void)
{
  printf("sizeof = %d\n",sizeof(struct vertex));
  int vplen = 2;
  struct vertex *vp= malloc_or_die (sizeof(struct vertex)* vplen);

  {
    int x = 2, y = 0; XY_ROTATE_PLUS120(x,y);
    if (x!=-1 || y != 1) { abort(); }
  }
  {
    int x = -1, y = 1; XY_ROTATE_PLUS120(x,y);
    if (x!=-1 || y != -1) { abort(); }
  }

  {
    int x = 2, y = 0; XY_ROTATE_PLUS60(x,y);
    if (x!=1 || y != 1) { abort(); }
  }
  {
    int x = 1, y = 1; XY_ROTATE_PLUS60(x,y);
    if (x!=-1 || y != 1) { abort(); }
  }

  vp[0].x = 0;
  vp[0].y = 0;
  vp[0].num_neighbours = 1;
  vp[0].neighbours[0] = 1;

  vp[1].x = 2;
  vp[1].y = 0;
  vp[1].num_neighbours = 1;
  vp[1].neighbours[0] = 0;

  if (find_xy(vp,vplen,2,0) != vp+1) { abort(); }

  distance(vp,vplen, 0,vplen-1);

  /* unfold (&vp, &vplen); */
  show(vp,vplen);
  /* distance(vp,vplen, 0,vplen-1); */


  printf ("distances start to end\n");
  int level;
  for (level = 1; level < 20; level++) {
    unfold (&vp, &vplen);
    int d;

    /* plain */
    {
      /* if(level <= 3) { show(vp,vplen);  */
      d = distance(vp,vplen, 0,vplen-1);
    }

    /* struct vertex *twin_vp; */
    /* int twin_vplen; */
    /* twin (vp,vplen, &twin_vp,&twin_vplen); */
    /* if (level <= 3) { printf("\n"); show(twin_vp,twin_vplen); } */
    /* int d = distance(twin_vp,twin_vplen, 0,vplen-1);  /\* end of untwinned *\/ */
    /* free(twin_vp); */

    printf("%d,", d);
    fflush(stdout);
  }

  return 0;
}
